
local item = function(name, anim, x)
    graphics[name] = function(t)
        turn(0.5, 0, 0)
        x(t)
    end
    classes.galaxy[name] = 1
    if anim == 0 then
        graphics[name .. ".animation"] = "static"
    end
    if anim == 2 then
        graphics[name .. ".animation"] = "around" 
    end
end

pi = math.pi
local u = (1 + math.sqrt(5)) / 2
local s = math.sqrt(3)

-- 12 vertices and 20 triangule faces
icosahedron_coords = {
    {0, 1, u}, -- top B 1
    {0,-1, u}, -- top F 2
    {0, 1, -u},
    {0, -1, -u},
    {1, u, 0}, -- R B 5
    {-1, u, 0}, -- L B 6
    {1, -u, 0}, -- R F 7
    {-1, -u, 0}, -- L F 8
    {u, 0, 1}, -- top R 9
    {u, 0, -1},
    {-u, 0, 1}, -- top L 11
    {-u, 0, -1},
}

icosahedron_triangles = {
    {1, 11, 2}, {1, 6, 11}, {1, 5, 6}, {1, 9, 5}, {1, 2, 9}, -- top back
    {2, 7, 9}, {2, 8, 7}, {2, 11, 8}, -- top front

    {3, 10, 4}, {3, 5, 10}, {3, 6, 5}, {3, 12, 6}, {3, 4, 12}, -- bottom back
    {4, 10, 7}, {4, 7, 8}, {4, 8, 12}, -- bottom front

    {11, 12, 8}, {12, 11, 6}, -- left
    {9, 7, 10}, {10, 5, 9}, -- right
    }

-- 20 vertices and 12 pentagon faces
dodecahedron_coords = {
    {-1, -1, -1},
    {-1, -1,  1},
    {-1,  1, -1},
    {-1,  1,  1},
    { 1, -1, -1},
    { 1, -1,  1},
    { 1,  1, -1},
    { 1,  1,  1},
    { 0, -1 / u, -u},
    { 0, -1 / u,  u},
    { 0,  1 / u, -u},
    { 0,  1 / u,  u},
    { -1 / u, -u, 0},
    { -1 / u,  u, 0},
    {  1 / u, -u, 0},
    {  1 / u,  u, 0},
    { -u, 0, -1 / u},
    {  u, 0, -1 / u},
    { -u, 0, 1 / u},
    {  u, 0, 1 / u},
}

local function _device2(t)

    turn(0, math.sin(t * math.pi) * 0.1, 0)

    move(0, 7, 0)
    scale(5)
    
    color("gray")
    make("sphere")

    color("orange")
    for i, coord in pairs(icosahedron_coords) do
        push()
        x, y, z = table.unpack(coord)
        lookto(x, y, z)
        move(0, 0, .9 + 0.1 * math.sin(t * 2 * math.pi))
        size(0.2, 0.2, 0.2)
        move(0, 0, 1)
        make("cylinder")
        pop()
    end

    x = 1 / math.sqrt(3)
    turn(x, x, x, 0.2)

    color("yellow")
    for i, coord in pairs(dodecahedron_coords) do
        push()
        x, y, z = table.unpack(coord)
        lookto(x, y, z)
        move(0, 0, .9 - 0.1 * math.sin(t * 2 * math.pi))
        size(0.2, 0.2, 0.2)
        move(0, 0, 1)
        make("cylinder")
        pop()
    end
    
end

item("device2_floating", 2, function(t)
    size(0.1)
    _device2(t)
end)

