#include "brick.h"
Brick* get_brick(int x, int y) {
    Application * app = app_global();
    Brick * b = NULL;
    int s = app->world->s;
    int ix, iy;
    pos_to_hex(x, y, & ix, & iy);
    ix = ix & (s - 1);
    iy = iy & (s - 1);
    {
        LandArrayIterator __iter0__ = LandArrayIterator_first(app->world->bricks);
        for (Brick * already = LandArrayIterator_item(app->world->bricks, &__iter0__); LandArrayIterator_next(app->world->bricks, &__iter0__); already = LandArrayIterator_item(app->world->bricks, &__iter0__)) {
            if (already->x == ix && already->y == iy) {
                if (b && b->p.z > already->p.z) {
                    continue;
                }
                b = already;
            }
        }
    }
    return b;
}
void add_brick(int kind, int x, int y) {
    float z = get_top_z(x, y);
    add_brick_z(kind, x, y, z);
}
void add_brick_offset(int kind, int x, int y, int z) {
    float z1 = get_top_z(x, y);
    add_brick_z(kind, x, y, z1 + z);
}
void add_brick_z(int kind, int x, int y, int z) {
    Application * app = app_global();
    Brick * b;
    land_alloc(b);
    b->kind = kind;
    x = wrap(x, app->world->landscape->w);
    y = wrap(y, app->world->landscape->w);
    b->p.x = x;
    b->p.y = y;
    land_array_add(app->world->bricks, b);
    int s = app->world->s;
    int ix, iy;
    pos_to_hex(x, y, & ix, & iy);
    int iz = floor(z / 4);
    if (iz < 0) {
        iz = 0;
    }
    if (iz > 31) {
        iz = 31;
    }
    ix = ix & (s - 1);
    iy = iy & (s - 1);
    b->x = ix;
    b->y = iy;
    b->z = iz;
    b->p.z = iz * 4;
    app->world->bricks_map [iz * s * s + iy * s + ix] = kind;
}
void del_brick(int x, int y) {
    Application * app = app_global();
    int s = app->world->s;
    Brick * b = get_brick(x, y);
    if (b) {
        if (b->kind == Mark || b->kind == Goal) {
            return ;
        }
        int i = - 1;
        {
            LandArrayIterator __iter0__ = LandArrayIterator_first(app->world->bricks);
            for (Brick * already = LandArrayIterator_item(app->world->bricks, &__iter0__); LandArrayIterator_next(app->world->bricks, &__iter0__); already = LandArrayIterator_item(app->world->bricks, &__iter0__)) {
                i++;
                if (already == b) {
                    break;
                }
            }
        }
        if (i >= 0) {
            land_array_remove(app->world->bricks, i);
        }
        app->world->bricks_map [b->z * s * s + b->y * s + b->x] = Nothing;
    }
}
void brick_draw_offset(Brick * b, Land4x4Matrix cm, Camera * cam, float ox, float oy) {
    Application * app = app_global();
    Land4x4Matrix matrix = land_4x4_matrix_mul(cm, land_4x4_matrix_translate(b->p.x + ox, b->p.y + oy, b->p.z));
    if (b->kind == Wall) {
        if (check_pos(b->x, b->y, b->z + 1) != Wall) {
            model_render_object(app->world->wall [0], matrix, cam);
        }
        if (1) {
            //check_pos(b.x - 1, b.y, b.z) != Wall:
            model_render_object(app->world->wall [1], matrix, cam);
        }
        if (1) {
            // check_pos(b.x, b.y - 1, b.z) != Wall:
            model_render_object(app->world->wall [2], matrix, cam);
        }
        if (1) {
            // check_pos(b.x + 1, b.y, b.z) != Wall:
            model_render_object(app->world->wall [3], matrix, cam);
        }
        if (1) {
            // check_pos(b.x, b.y + 1, b.z) != Wall:
            model_render_object(app->world->wall [4], matrix, cam);
        }
        if (1) {
            model_render_object(app->world->wall [5], matrix, cam);
        }
        if (1) {
            model_render_object(app->world->wall [6], matrix, cam);
        }
    }
    if (b->kind == Bridge) {
        model_render_object(app->world->bridge, matrix, cam);
    }
    if (b->kind == Mark) {
        Camera cam2;
        camera_init(& cam2);
        cam2.p = b->p;
        draw_model(app->model_mark, & cam2, 1, app->light);
    }
    if (b->kind == Goal) {
        Camera cam2;
        camera_init(& cam2);
        cam2.p = b->p;
        draw_model(app->model_goal, & cam2, 1, app->light);
    }
}
void bricks_draw(Camera * cam, float ox, float oy) {
    Application * app = app_global();
    Land4x4Matrix cm = camera_matrix(cam);
    {
        LandArrayIterator __iter0__ = LandArrayIterator_first(app->world->bricks);
        for (Brick * b = LandArrayIterator_item(app->world->bricks, &__iter0__); LandArrayIterator_next(app->world->bricks, &__iter0__); b = LandArrayIterator_item(app->world->bricks, &__iter0__)) {
            brick_draw_offset(b, cm, cam, ox, oy);
        }
    }
}
void trans(Land4x4Matrix * t, Land4x4Matrix m) {
    * t = land_4x4_matrix_mul(* t, m);
}
void make_wall_hex(World * w) {
    // 0..7 = top, back-left, back, back-right, front-right, front, front-left, bottom
    Material * m = material_new(1, 1, 1, 1);
    w->wall [0] = mesh_new();
    if (1) {
        Object * o = object_regular(6, m);
        Land4x4Matrix transform = land_4x4_matrix_identity();
        trans(& transform, land_4x4_matrix_translate(0, 0, 4));
        trans(& transform, land_4x4_matrix_scale(2.75, 2 / sqrt(0.75), 1));
        object_transform(o, & transform);
        mesh_add_object(w->wall [0], o);
    }
    w->wall [7] = mesh_new();
    for (int i = 0; i < 6; i += 1) {
        Object * o = object_regular(4, m);
        Land4x4Matrix transform = land_4x4_matrix_identity();
        trans(& transform, land_4x4_matrix_scale(2.75, 2 / sqrt(0.75), 4));
        trans(& transform, land_4x4_matrix_rotate(0, 0, 1, (1 - i) * pi / 3));
        trans(& transform, land_4x4_matrix_translate(0, sqrt(0.75), 0.5));
        trans(& transform, land_4x4_matrix_rotate(1, 0, 0, - pi / 2));
        trans(& transform, land_4x4_matrix_rotate(0, 0, 1, pi / 4));
        trans(& transform, land_4x4_matrix_scale(1 / sqrt(2), 1 / sqrt(2), 1));
        object_transform(o, & transform);
        w->wall [1 + i] = mesh_new();
        mesh_add_object(w->wall [1 + i], o);
    }
}
void make_bridge_hex(World * w) {
    Material * m = material_new(1, 1, 0, 1);
    w->bridge = mesh_new();
    Object * o = object_prism(6, m);
    Land4x4Matrix transform = land_4x4_matrix_identity();
    trans(& transform, land_4x4_matrix_translate(0, 0, 4));
    trans(& transform, land_4x4_matrix_scale(2.75, 2 / sqrt(0.75), 0.25));
    trans(& transform, land_4x4_matrix_rotate(0, 0, 1, pi / 2));
    trans(& transform, land_4x4_matrix_translate(0, 0, - 1));
    object_transform(o, & transform);
    mesh_add_object(w->bridge, o);
}
int check_pos(int x, int y, int z) {
    Application * app = app_global();
    int s = app->world->s;
    x &= s - 1;
    y &= s - 1;
    return app->world->bricks_map [z * s * s + y * s + x];
}
int check_wall_z(int x, int y, int z) {
    int b = check_pos(x, y, z);
    if (b == Wall) {
        return z * 4 + 4;
    }
    b = check_pos(x, y, z + 1);
    if (b == Wall) {
        return (z + 1) * 4 + 4;
    }
    if (z > 0) {
        b = check_pos(x, y, z - 1);
        if (b == Wall) {
            return (z - 1) * 4 + 4;
        }
    }
    return 0;
}
int check_bridge_z(int x, int y, int z) {
    int b = check_pos(x, y, z);
    if (b == Bridge) {
        return z * 4 + 4;
    }
    b = check_pos(x, y, z + 1);
    if (b == Bridge) {
        return (z + 1) * 4 + 4;
    }
    if (z > 0) {
        b = check_pos(x, y, z - 1);
        if (b == Bridge) {
            return (z - 1) * 4 + 4;
        }
    }
    return 0;
}
int find_highest_z(int x, int y) {
    for (int z = 31; z > - 1; z += - 1) {
        int kind = check_pos(x, y, z);
        if (kind == Wall || kind == Bridge) {
            return z + 1;
        }
    }
    return 0;
}
void build_bridge(float px, float py) {
    Application * app = app_global();
    int hx, hy;
    pos_to_hex(px, py, & hx, & hy);
    int hz = find_highest_z(hx, hy);
    if (hz > 0) {
        if (check_pos(hx, hy, hz - 1)) {
            return ;
        }
    }
    for (int d = 0; d < 6; d += 1) {
        int x2 = hx;
        int y2 = hy;
        add_dir(& x2, & y2, d);
        int z = find_highest_z(x2, y2);
        if (z > hz) {
            hz = z;
        }
    }
    float lz = landscape_get_height_interpolated(app->world->landscape, px, py);
    if (lz > hz * 4) {
        hz = floor(lz / 4);
    }
    else if (hz > 0) {
        hz -= 1;
    }
    add_brick_z(Bridge, px, py, hz * 4);
}
void clear_bricks(void) {
    Application * app = app_global();
    {
        LandArrayIterator __iter0__ = LandArrayIterator_first(app->world->bricks);
        for (Brick * already = LandArrayIterator_item(app->world->bricks, &__iter0__); LandArrayIterator_next(app->world->bricks, &__iter0__); already = LandArrayIterator_item(app->world->bricks, &__iter0__)) {
            land_free(already);
        }
    }
    land_array_destroy(app->world->bricks);
}
