#include "game.h"
static bool Debug = 0;
static void reset_camera(void);
static void _hex2(float ux, float uy, float r, float g, float b, float a, float t, float z);
static void _hex(float ux, float uy, float r, float g, float b, float a, float t);
static void next_song(void);
static void _dialog(int start, int end, str text, int style);
static void reset_camera(void) {
    Application * app = app_global();
    camera_init(& app->cam);
    double angle = - pi * 0.33;
    app->cam.y = land_vector_rotate(app->cam.y, app->cam.x, angle);
    app->cam.z = land_vector_rotate(app->cam.z, app->cam.x, angle);
    app->cam.zoom = 2;
    app->cam.p.x = app->start_x;
    app->cam.p.y = app->start_y;
    app->zoom_to = 3.5;
}
void game_init(void) {
    Application * app = app_global();
    app->music = 1;
    app->stream = land_stream_new(2048, 4, 22050, 16, 2);
    next_song();
    app->win = land_sound_load("sound/win.ogg");
    app->yell = land_sound_load("sound/yell.ogg");
    app->subdivisions = 1;
    app->light = land_vector(0.75, - 1, 2);
    app->units = land_array_new();
    graphics_load_all();
    app->model_allefant = script_graphics_load("allefant");
    if (! app->model_allefant->frames->count) {
        print("Could not load data");
        abort();
    }
    app->model_allefant->scale = 3;
    app->model_mothership = script_graphics_load("shipc4");
    app->model_mothership->scale = 8;
    app->model_mark = script_graphics_load("mark");
    app->model_mark->scale = 4;
    app->model_goal = script_graphics_load("device2_floating");
    app->model_goal->scale = 4;
    app->seed = land_rand(0, 1000000000);
    app->first = 1;
    game_new();
}
void game_new(void) {
    Application * app = app_global();
    app->ticks = 0;
    if (app->world) {
        world_destroy(app->world);
    }
    world_new();
    make_stars();
    reset_camera();
    app->done = 0;
    app->paused = 1;
    app->warmup = 1;
    app->released = 0;
    app->retrieved = 0;
    app->release = 0;
    app->died = 0;
}
void repeat_level(void) {
    Application * app = app_global();
    app->first = 0;
    game_new();
}
void next_level(void) {
    Application * app = app_global();
    app->first = 0;
    app->seed = land_rand(0, 1000000000);
    app->level++;
    game_new();
}
void game_tick(void) {
    Application * app = app_global();
    if (land_closebutton()) {
        land_quit();
    }
    if (land_key_pressed(LandKeyEscape)) {
        land_quit();
    }
    float dx = 0;
    float dy = 0;
    float dz = land_mouse_delta_z();
    float mx = land_mouse_x();
    float my = land_mouse_y();
    app->cam.zoom += dz * 0.1;
    if (app->cam.zoom > 5) {
        app->cam.zoom = 5;
    }
    if (app->cam.zoom < 2) {
        app->cam.zoom = 2;
    }
    if (app->warmup) {
        if (app->cam.zoom < app->zoom_to) {
            app->cam.zoom += 0.005;
        }
        else {
            app->warmup = 0;
            app->paused = 0;
        }
    }
    if (app->music && ! land_stream_is_playing(app->stream)) {
        next_song();
    }
    if (land_key_pressed('m')) {
        app->music = ! app->music;
        if (app->music) {
            land_stream_set_playing(app->stream, 1);
        }
        else {
            land_stream_set_playing(app->stream, 0);
        }
    }
    if (land_key(LandKeyLeft)) {
        dx = 1;
    }
    if (land_key(LandKeyRight)) {
        dx = - 1;
    }
    if (land_key(LandKeyUp)) {
        dy = 1;
    }
    if (land_key(LandKeyDown)) {
        dy = - 1;
    }
    if (land_key_pressed('n')) {
        app->level = 0;
        app->first = 1;
        app->seed = land_rand(0, 1000000000);
        game_new();
    }
    if (Debug) {
        if (land_key_pressed('r')) {
            app->debug_wireframe = ! app->debug_wireframe;
        }
        if (land_key_pressed('l')) {
            app->debug_no_landscape = ! app->debug_no_landscape;
        }
        if (land_key_pressed('u')) {
            app->debug_no_units = ! app->debug_no_units;
        }
        if (land_key_pressed('g')) {
            app->debug_grid = ! app->debug_grid;
        }
        if (land_key_pressed('c')) {
            app->retrieved = app->elephants;
        }
        if (land_key_pressed('d')) {
            app->died = app->elephants;
        }
    }
    if (land_key_pressed('p')) {
        app->paused = ! app->paused;
    }
    if (land_key_pressed('e') && Debug) {
        Unit * u = unit_new(UnitAllefant, app->under.x, app->under.y);
        camera_change_freely(& u->cam, 0, 0, land_rnd(0, pi * 2));
        u->cam.p.z = 30;
    }
    LandVector v = world_get_position_under_z(& app->cam, mx, my);
    app->under = v;
    bool scroll = 0;
    bool turn = 0;
    if (land_mouse_button(0)) {
        if (land_key(LandKeyShift) || land_mouse_button(1)) {
            scroll = 1;
        }
    }
    else if (land_mouse_button(1)) {
        turn = 1;
    }
    if (land_mouse_button(2)) {
        scroll = 1;
    }
    int ux, uy;
    pos_to_hex(v.x, v.y, & ux, & uy);
    hex_to_int(ux, uy, & app->ux, & app->uy);
    if (app->ux != app->ux_ || app->uy != app->uy_) {
        app->ux_ = app->ux;
        app->uy_ = app->uy;
        app->can_build = 1;
    }
    if (land_mouse_button(0) && ! scroll && ! turn) {
        if (app->can_build) {
            if (app->selection == 0) {
                app->show_help = 1;
            }
            else {
                app->show_help = 0;
            }
            if (app->selection == 1) {
                build_bridge(app->ux, app->uy);
            }
            if (app->selection == 2) {
                float z = get_top_z(app->ux, app->uy);
                if (z >= 0.01) {
                    add_brick(Wall, app->ux, app->uy);
                    app->can_build = 0;
                }
            }
        }
    }
    else {
        app->can_build = 1;
    }
    if (land_key_pressed('1')) {
        app->selection = 1;
    }
    if (land_key_pressed('2')) {
        app->selection = 2;
    }
    if (land_key_pressed(' ')) {
        if (app->selection == 2) {
            app->selection = 1;
        }
        else {
            app->selection = 2;
        }
    }
    if (land_key_pressed(LandKeyDelete)) {
        del_brick(app->ux, app->uy);
    }
    camera_move(& app->cam, - dx, dy, 0);
    app->cam.p.x = wrap(app->cam.p.x, app->world->landscape->w);
    app->cam.p.y = wrap(app->cam.p.y, app->world->landscape->w);
    if (turn) {
        dx = land_mouse_delta_x() * 0.01;
        dy = land_mouse_delta_y() * 0.01;
        camera_change_locked_constrained(& app->cam, dy, - dx, 0.1, 1.0);
    }
    if (scroll) {
        dx = land_mouse_delta_x() * 1.0 * pow(2, - app->cam.zoom);
        dy = land_mouse_delta_y() * 1.0 * pow(2, - app->cam.zoom);
        camera_move(& app->cam, - dx, dy, 0);
        land_vector_iadd(& app->starpos, land_vector_mul(app->cam.x, - dx * 0.1));
        land_vector_iadd(& app->starpos, land_vector_mul(app->cam.y, dy * 0.1));
    }
    if (! app->paused) {
        if (tick_intro()) {
            return ;
        }
        units_tick();
        app->ticks++;
    }
}
LandVector get_point(Camera * cam, int x, int y, float z) {
    float w = land_display_width();
    float h = land_display_height();
    LandFloat s = pow(2, cam->zoom);
    LandVector v = land_vector(x, y, z);
    v = land_vector_transform(v, cam->p, cam->x, cam->y, cam->z);
    v.x *= s;
    v.y *= s;
    v.x = v.x + w / 2;
    v.y = h / 2 - v.y;
    return v;
}
LandVector get_point2(Camera * cam, Landscape * ls, float x, float y) {
    float w = land_display_width();
    float h = land_display_height();
    LandFloat s = pow(2, cam->zoom);
    float z = landscape_get_height_interpolated(ls, x, y);
    LandVector v = land_vector(x, y, z);
    v = land_vector_transform(v, cam->p, cam->x, cam->y, cam->z);
    v.x *= s;
    v.y *= s;
    v.x = v.x + w / 2;
    v.y = h / 2 - v.y;
    v.z = z;
    return v;
}
static void _hex2(float ux, float uy, float r, float g, float b, float a, float t, float z) {
    Application * app = app_global();
    Camera * cam = & app->cam;
    int ix, iy, lx, ly;
    pos_to_hex(ux, uy, & ix, & iy);
    hex_to_int(ix, iy, & lx, & ly);
    LandVector v [7];
    v [0] = get_point(cam, lx - 1.25, ly - 2, z);
    v [1] = get_point(cam, lx + 1.25, ly - 2, z);
    v [2] = get_point(cam, lx + 2.75, ly, z);
    v [3] = get_point(cam, lx + 1.25, ly + 2, z);
    v [4] = get_point(cam, lx - 1.25, ly + 2, z);
    v [5] = get_point(cam, lx - 2.75, ly, z);
    land_premul(r, g, b, a);
    land_thickness(t);
    for (int i = 0; i < 6; i += 1) {
        land_line(v [i].x, v [i].y, v [(i + 1) % 6].x, v [(i + 1) % 6].y);
    }
}
static void _hex(float ux, float uy, float r, float g, float b, float a, float t) {
    Application * app = app_global();
    Camera * cam = & app->cam;
    World * world = app->world;
    Landscape * l = world->landscape;
    int ix, iy, lx, ly;
    pos_to_hex(ux, uy, & ix, & iy);
    hex_to_int(ix, iy, & lx, & ly);
    LandVector v [7];
    v [0] = get_point2(cam, l, lx - 1.25, ly - 2);
    v [1] = get_point2(cam, l, lx + 1.25, ly - 2);
    v [2] = get_point2(cam, l, lx + 2.75, ly);
    v [3] = get_point2(cam, l, lx + 1.25, ly + 2);
    v [4] = get_point2(cam, l, lx - 1.25, ly + 2);
    v [5] = get_point2(cam, l, lx - 2.75, ly);
    if (v [0].z < 0.01 && v [1].z < 0.01 && v [2].z < 0.01 && v [3].z < 0.01 && v [4].z < 0.01 && v [5].z < 0.01) {
        land_premul(0, 0, 1, a);
    }
    else {
        land_premul(r, g, b, a);
    }
    land_thickness(t);
    for (int i = 0; i < 6; i += 1) {
        land_line(v [i].x, v [i].y, v [(i + 1) % 6].x, v [(i + 1) % 6].y);
    }
}
void draw_under(void) {
    Application * app = app_global();
    LandVector u = app->under;
    int ix, iy, lx, ly;
    pos_to_hex(u.x, u.y, & ix, & iy);
    hex_to_int(ix, iy, & lx, & ly);
    int x2 [] = {- 8, - 4, 0, 4, 8, 8, 8, 4, 0, - 4, - 8, - 8};
    int y2 [] = {4, 6, 8, 6, 4, 0, - 4, - 6, - 8, - 6, - 4, 0};
    for (int i = 0; i < 12; i += 1) {
        _hex(lx + x2 [i], ly + y2 [i], 1, 1, 0, 0.5, 1);
    }
    int x1 [] = {- 4, 0, 4, 4, 0, - 4};
    int y1 [] = {2, 4, 2, - 2, - 4, - 2};
    for (int i = 0; i < 6; i += 1) {
        _hex(lx + x1 [i], ly + y1 [i], 1, 1, 0, 0.7, 2);
    }
    Brick * b = get_brick(lx, ly);
    if (b) {
        _hex2(lx, ly, 1, 0, 0, 0.9, 3, b->p.z + 4);
    }
    else {
        _hex(lx, ly, 1, 1, 0, 0.9, 3);
    }
}
void draw_debug_grid(void) {
    Application * app = app_global();
    World * world = app->world;
    land_reset_transform();
    land_display_set_default_shaders();
    glDisable(GL_CULL_FACE);
    land_render_state(LAND_WRITE_MASK, LAND_RGBA_MASK);
    land_render_state(LAND_DEPTH_TEST, 0);
    landscape_draw_grid(world->landscape, & app->cam);
}
void game_draw(void) {
    Application * app = app_global();
    World * world = app->world;
    land_clear(0, 0, 0, 1);
    glEnable(GL_CULL_FACE);
    land_render_state(LAND_DEPTH_TEST, 1);
    land_render_state(LAND_WRITE_MASK, LAND_RGBA_MASK + LAND_DEPTH_MASK);
    land_clear_depth(1);
    if (1) {
        Camera cam2 = app->cam;
        cam2.p = app->starpos;
        Land4x4Matrix matrix = camera_matrix(& cam2);
        LandVector light = land_vector(cam2.z.x, cam2.z.y, cam2.z.z);
        model_render_object_light2(app->stars, matrix, world_get_projection_matrix(), & app->cam, light);
    }
    world_render(world, & app->cam);
    if (! app->debug_no_units) {
        float s = world->landscape->w;
        units_draw_offset(0, 0);
        float xo = app->cam.p.x < s / 2 ? - s : s;
        float yo = app->cam.p.y < s / 2 ? - s : s;
        units_draw_offset(xo, 0);
        units_draw_offset(xo, yo);
        units_draw_offset(0, yo);
    }
    land_reset_transform();
    render_switch_shader(NULL);
    land_display_set_default_shaders();
    glDisable(GL_CULL_FACE);
    land_render_state(LAND_WRITE_MASK, LAND_RGBA_MASK);
    land_render_state(LAND_DEPTH_TEST, 0);
    if (app->debug_grid) {
        draw_debug_grid();
    }
    draw_under();
    double scale = land_scale_to_fit(1280, 720, 0);
    if (! app->font) {
        load_font(scale);
    }
    draw_intro();
}
void game_done(void) {
    ;
}
void make_stars(void) {
    Application * app = app_global();
    if (app->stars) {
        mesh_destroy(app->stars);
    }
    Mesh * mesh = mesh_new();
    for (int i = 0; i < 1000; i += 1) {
        float x = land_rnd(- 400, 400);
        float y = land_rnd(- 400, 400);
        float z = land_rnd(- 300, - 10);
        float s = 1 - (- z - 10) / (300.0 - 10);
        float c = 0.7 + s * 0.3;
        float r = c;
        float g = c;
        float b = land_rnd(0, c);
        s = 0.1 + s * 2;
        Material * mat = material_new(r, g, b, 1);
        Object * o = object_icosphere(1, mat);
        Land4x4Matrix transform = land_4x4_matrix_identity();
        trans(& transform, land_4x4_matrix_translate(x, y, z));
        trans(& transform, land_4x4_matrix_scale(s, s, s));
        object_transform(o, & transform);
        mesh_add_object(mesh, o);
    }
    app->stars = mesh;
}
static void next_song(void) {
    str songs [] = {"music/marching_synth.ogg", "music/marching_western.ogg", "music/marching_island.ogg", "music/marching_eastern.ogg", "music/marching_singing.ogg"} /* easteregg: hear my cringy recording :P */;
    int order [15] = {0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 2, 3, 0, 1, 4};
    Application * app = app_global();
    char const * name = NULL;
    name = songs [order [app->song]];
    app->song++;
    app->song %= 15;
    land_stream_music_once(app->stream, name);
    land_stream_volume(app->stream, 0.5);
}
void load_font(double scale) {
    Application * app = app_global();
    if (app->font) {
        land_font_destroy(app->font);
    }
    double s = scale;
    app->font = land_font_load("fonts/URWGothicBook.ttf", 32 * s);
    land_font_scale(app->font, 1.0 / s);
}
static void _dialog(int start, int end, str text, int style) {
    Application * app = app_global();
    start *= 60;
    end *= 60;
    land_color(1, 1, 1, 1);
    int b = ((int)(app->ticks / 15)) % 2;
    if (style == 1) {
        land_color(b, 0, 0, b);
    }
    land_text_pos(640, 180);
    if (app->ticks >= start && app->ticks < end) {
        land_print_middle(text);
    }
}
bool tick_intro(void) {
    Application * app = app_global();
    if (app->done) {
        if (app->ticks == 5 * 60) {
            if (app->retrieved >= app->elephants) {
                next_level();
            }
            else {
                repeat_level();
            }
            return 1;
        }
        return 0;
    }
    if (app->ticks == 0) {
        add_brick_offset(Goal, app->goal_x, app->goal_y, 4);
    }
    if (app->first) {
        if (app->ticks == 17 * 60) {
            add_brick_offset(Mark, app->start_x, app->start_y, 4);
            app->release = 1;
        }
        if (app->ticks == 25 * 60) {
            app->first = 0;
        }
    }
    else {
        if (app->ticks == 0) {
            app->release = 1;
            add_brick_offset(Mark, app->start_x, app->start_y, 4);
        }
    }
    if (app->retrieved + app->died >= app->elephants) {
        app->done = 1;
        app->ticks = 0;
        return 1;
    }
    return 0;
}
void draw_intro(void) {
    Application * app = app_global();
    if (app->show_help) {
        land_text_pos(0, 720);
        land_color(1, 1, 1, 1);
        land_print_bottom("Space/1/2 to toggle/select Bridge/Wall");
    }
    if (app->done) {
        if (app->retrieved >= app->elephants) {
            _dialog(1, 5, "All elephants secured!", 0);
        }
        else if (app->died >= app->elephants) {
            _dialog(1, 2, "Uh, I think...", 0);
            _dialog(2, 5, "Some didn't make it...", 1);
        }
        else {
            _dialog(1, 2, "All elephants...", 0);
            _dialog(2, 5, "I mean some elephants secured!", 0);
        }
        return ;
    }
    if (app->first) {
        _dialog(1, 3, "Left!", 0);
        _dialog(4, 7, "I said LEFT!", 0);
        _dialog(8, 11, "I mean right", 0);
        _dialog(12, 15, "go right!", 0);
        _dialog(17, 25, "Release the elephants!", 1);
        if (app->ticks < 25 * 60) {
            return ;
        }
    }
    else {
        if (app->level == 0) {
            _dialog(1, 2, "Here we go again...", 0);
            _dialog(2, 5, "Release the elephants!", 1);
        }
        else {
            char s [100];
            sprintf(s, "Level %d!", 1 + app->level);
            _dialog(1, 2, s, 0);
            if (app->level == 1) {
                _dialog(2, 5, "Check out Allefant 1!", 1);
            }
            else {
                _dialog(2, 5, "Release the elephants!", 1);
            }
        }
    }
    land_text_pos(0, 0);
    land_color(1, 0.8, 0, 1);
    land_print("%d/%d", app->retrieved, app->released);
}
