#include "game/camera.h"
static void _test(str name, Camera cam, double result, double expected);
char* camera_to_string(Camera * self) {
    /* Free the returned string with land_free!
     */
    LandBuffer * b = land_buffer_new();
    Vector * p = & self->p;
    Vector * x = & self->x;
    Vector * y = & self->y;
    Vector * z = & self->z;
    land_buffer_addf(b, "%.3f %.3f %.3f ", p->x, p->y, p->z);
    land_buffer_addf(b, "%.3f %.3f %.3f ", x->x, x->y, x->z);
    land_buffer_addf(b, "%.3f %.3f %.3f ", y->x, y->y, y->z);
    land_buffer_addf(b, "%.3f %.3f %.3f ", z->x, z->y, z->z);
    land_buffer_addf(b, "%.3f", self->zoom);
    return land_buffer_finish(b);
}
void camera_from_string(Camera * self, char const * s) {
    Vector * p = & self->p;
    Vector * x = & self->x;
    Vector * y = & self->y;
    Vector * z = & self->z;
    sscanf(s, "%lg %lg %lg ""%lg %lg %lg ""%lg %lg %lg ""%lg %lg %lg""%lg", & p->x, & p->y, & p->z, & x->x, & x->y, & x->z, & y->x, & y->y, & y->z, & z->x, & z->y, & z->z, & self->zoom);
}
Camera* camera_new(void) {
    Camera * self;
    land_alloc(self);
    camera_init(self);
    return self;
}
void camera_init(Camera * c) {
    memset(c, 0, sizeof (* c));
    c->x.x = 1;
    c->y.y = 1;
    c->z.z = 1;
}
Camera camera_identity(void) {
    Camera c;
    camera_init(& c);
    return c;
}
void camera_change_freely(Camera * self, float x, float y, float z) {
    Vector xaxis = self->x;
    self->x = land_vector_rotate(self->x, xaxis, x);
    self->y = land_vector_rotate(self->y, xaxis, x);
    self->z = land_vector_rotate(self->z, xaxis, x);
    Vector yaxis = self->y;
    self->x = land_vector_rotate(self->x, yaxis, y);
    self->y = land_vector_rotate(self->y, yaxis, y);
    self->z = land_vector_rotate(self->z, yaxis, y);
    Vector zaxis = self->z;
    self->x = land_vector_rotate(self->x, zaxis, z);
    self->y = land_vector_rotate(self->y, zaxis, z);
    self->z = land_vector_rotate(self->z, zaxis, z);
}
void camera_change_locked(Camera * self, float x, float z) {
    Vector xaxis = self->x;
    self->x = land_vector_rotate(self->x, xaxis, x);
    self->y = land_vector_rotate(self->y, xaxis, x);
    self->z = land_vector_rotate(self->z, xaxis, x);
    Vector zaxis = land_vector(0, 0, 1);
    self->x = land_vector_rotate(self->x, zaxis, z);
    self->y = land_vector_rotate(self->y, zaxis, z);
    self->z = land_vector_rotate(self->z, zaxis, z);
}
void camera_change_locked_constrained(Camera * self, float x, float z, float min_x, float max_x) {
    Vector zaxis = land_vector(0, 0, 1);
    LandVector c = land_vector_cross(self->z, zaxis);
    double angle = asin(1 - land_vector_norm(c));
    if (angle + x < min_x) {
        x = min_x - angle;
    }
    if (angle + x > max_x) {
        x = max_x - angle;
    }
    Vector xaxis = self->x;
    self->x = land_vector_rotate(self->x, xaxis, x);
    self->y = land_vector_rotate(self->y, xaxis, x);
    self->z = land_vector_rotate(self->z, xaxis, x);
    self->x = land_vector_rotate(self->x, zaxis, z);
    self->y = land_vector_rotate(self->y, zaxis, z);
    self->z = land_vector_rotate(self->z, zaxis, z);
}
LandFloat camera_get_pitch(Camera * self) {
    /* get the rotation around x (right)
     * Z
     * |   y
     * |  /.  y.z
     * | / .
     * |/__._____Y
     * X
     */
    LandFloat zy = sqrt(self->y.z * self->y.z + self->y.y * self->y.y);
    return - asin(self->y.z / zy);
}
LandFloat camera_get_roll(Camera * self) {
    /* get rotation around y (back)
     * X
     * | x.z
     * |....x
     * |  /
     * |/________Z
     * Y
     */
    LandFloat xz = sqrt(self->x.x * self->x.x + self->x.z * self->x.z);
    return asin(self->x.z / xz);
}
LandFloat camera_get_yaw(Camera * self) {
    /* get rotation around z (up)
     * Y
     * |   x
     * |  /.  x.y
     * | / .
     * |/__._____X
     * Z
     * A return of 0 means looking along 1/0/0. A return of pi/2 means
     * looking along 0/1/0.
     */
    return - atan2(self->x.y, self->x.x);
}
void camera_set_yaw(Camera * self, LandFloat a) {
    self->x = land_vector(cos(- a), sin(- a), 0);
    self->z = land_vector(0, 0, 1);
    // z = x × y
    // x = y × z
    // y = z × x
    self->y = land_vector_cross(self->z, self->x);
}
void camera_move(Camera * self, float x, float y, float z) {
    land_vector_iadd(& self->p, land_vector_mul(self->x, x));
    LandVector up = land_vector(0, 0, 1);
    LandVector back = land_vector_cross(self->x, up);
    back = land_vector_normalize(back);
    land_vector_iadd(& self->p, land_vector_mul(back, - y));
    land_vector_iadd(& self->p, land_vector_mul(up, z));
}
void camera_shift(Camera * self, float x, float y) {
    LandVector xv = land_vector_normalize(land_vector(self->x.x, self->x.y, 0));
    LandVector yv = land_vector_normalize(land_vector(self->y.x, self->y.y, 0));
    yv = land_vector_mul(yv, 1 + sin(fabs(self->y.z)));
    land_vector_iadd(& self->p, land_vector_mul(xv, x));
    land_vector_iadd(& self->p, land_vector_mul(yv, y));
}
void camera_look_into(Camera * self, float dx, float dy) {
    /* Change the camera to look into the given dx/dy direction, with
     * z going straight up.
     */
    double d = sqrt(dx * dx + dy * dy);
    if (d < 0.0001) {
        return ;
    }
    self->y = land_vector(dx / (- d), dy / (- d), 0);
    self->z = land_vector(0, 0, 1);
    self->x = land_vector_cross(self->y, self->z);
}
void camera_warp(Camera * self, float x, float y, float z) {
    self->p.x = x;
    self->p.y = y;
    self->p.z = z;
}
Land4x4Matrix camera_matrix(Camera * camera) {
    Land4x4Matrix m = land_4x4_matrix_inverse_from_vectors(& camera->p, & camera->x, & camera->y, & camera->z);
    // apply zoom
    LandFloat s = pow(2, camera->zoom);
    for (int i = 0; i < 8; i += 1) {
        m.v [i] *= s;
    }
    return m;
}
LandFloat camera_get_scale(Camera * camera) {
    return pow(2, camera->zoom);
}
Land4x4Matrix camera_forward_matrix(Camera * camera) {
    Land4x4Matrix m = land_4x4_matrix_from_vectors(& camera->p, & camera->x, & camera->y, & camera->z);
    return m;
}
Land4x4Matrix camera_forward_matrix_replace_position(Camera * camera, LandVector * pos) {
    Land4x4Matrix m = land_4x4_matrix_from_vectors(pos, & camera->x, & camera->y, & camera->z);
    return m;
}
// def camera_orientation(Camera *camera) -> Land4x4Matrix:
// Land4x4Matrix m = land_4x4_matrix_from_vectors(
// &camera.p, &camera.x, &camera.y, &camera.z)
// return m
void camera_init_from_matrix(Camera * camera, Land4x4Matrix matrix) {
    LandFloat * v = matrix.v;
    camera->x = land_vector_normalize(land_vector(v [0], v [1], v [2]));
    camera->y = land_vector_normalize(land_vector(v [4], v [5], v [6]));
    camera->z = land_vector_normalize(land_vector(v [8], v [9], v [10]));
    camera->p = land_vector(v [3], v [7], v [11]);
}
LandVector camera_get_on_screen_position(Camera * cam, LandVector pos) {
    LandFloat s = pow(2, cam->zoom);
    LandVector t = land_vector_transform(pos, cam->p, cam->x, cam->y, cam->z);
    return land_vector(t.x * s, t.y * s, 0);
}
static void _test(str name, Camera cam, double result, double expected) {
    double epsilon = 0.000001;
    double differ = fabs(result - expected);
    if (differ < epsilon) {
        print("OK %s", name);
    }
    else {
        print("FAIL %s", name);
        print("expected %f but got %f", expected, result);
        print("% .3f % .3f % .3f", cam.p.x, cam.p.y, cam.p.z);
        print("% .3f % .3f % .3f", cam.x.x, cam.x.y, cam.x.z);
        print("% .3f % .3f % .3f", cam.y.x, cam.y.y, cam.y.z);
        print("% .3f % .3f % .3f", cam.z.x, cam.z.y, cam.z.z);
    }
}
void camera_tests(void) {
    Camera cam;
    cam = camera_identity();
    _test("pitch", cam, camera_get_pitch(& cam), 0);
    camera_change_freely(& cam, pi / 2, 0, 0);
    _test("pitch2", cam, camera_get_pitch(& cam), pi / 2);
    camera_change_freely(& cam, pi, 0, 0);
    _test("pitch3", cam, camera_get_pitch(& cam), - pi / 2);
    cam = camera_identity();
    _test("roll", cam, camera_get_roll(& cam), 0);
    camera_change_freely(& cam, 0, pi / 2, 0);
    _test("roll2", cam, camera_get_roll(& cam), pi / 2);
    camera_change_freely(& cam, 0, pi, 0);
    _test("roll3", cam, camera_get_roll(& cam), - pi / 2);
    cam = camera_identity();
    for (int i = 0; i < 8; i += 1) {
        char yaw [10];
        sprintf(yaw, "yaw%d", 1 + i);
        float exp = 2 * pi * i / 8;
        if (exp >= pi) {
            exp -= 2 * pi;
        }
        _test(yaw, cam, camera_get_yaw(& cam), exp);
        camera_change_freely(& cam, 0, 0, 2 * pi / 8);
    }
}
