#include "game/isles.h"
typedef struct Isle Isle;
struct Isle {
    int x, y;
    int radius;
    int kind;
};
static void draw_isle(Landscape * landscape, Isle * isle);
static void update_light(Landscape * landscape, int tx, int ty);
static void draw_isle(Landscape * landscape, Isle * isle) {
    /* x^2 + y^2 + z^2 = s^2 -> z^2
     * -> s^2 - x^2 - y^2
     * r^2 + (s - h)^2 = s^2
     * -> r^2 + s^2 + h^2 - 2sh = s^2
     * -> r^2 + h^2 - 2sh = 0
     * -> s = (r^2 + h^2) / 2h
     */
    int cx = isle->x;
    int cy = isle->y;
    int r = isle->radius;
    double h = 4;
    double s = (r * r + h * h) / 2.0 / h;
    for (int y = - r; y < r + 1; y += 1) {
        for (int x = - r; x < r + 1; x += 1) {
            if (x * x + y * y > s * s) {
                continue;
            }
            double z = sqrt(s * s - x * x - y * y) - s + h;
            if (z < 0) {
                continue;
            }
            Tile * t = landscape_get_tile(landscape, cx + x, cy + y);
            if (z > t->height) {
                float d = z - t->height;
                t->height = z;
                if (t->kind < 1) {
                    t->kind = isle->kind;
                }
                else {
                    if (isle->kind == 1) {
                        t->kind -= d / 2;
                        if (t->kind < 1) {
                            t->kind = 1;
                        }
                    }
                    else if (isle->kind == 2) {
                        t->kind += d / 2;
                        if (t->kind > 2) {
                            t->kind = 2;
                        }
                    }
                }
            }
        }
    }
}
static void update_light(Landscape * landscape, int tx, int ty) {
    Tile * t = landscape_get_tile(landscape, tx, ty);
    Tile * nt [5];
    int dx [5] = {1, 0, - 1, 0, 1};
    int dy [5] = {0, 1, 0, - 1, 0};
    for (int i = 0; i < 4; i += 1) {
        nt [i] = landscape_get_tile(landscape, tx + dx [i], ty + dy [i]);
    }
    nt [4] = nt [0];
    double v = 1.0 / sqrt(3);
    LandVector light = land_vector(v, v, v);
    float nx = 0, ny = 0, nz = 0;
    float l = 0;
    for (int i = 0; i < 4; i += 1) {
        LandVector d0 = land_vector(dx [i], dy [i], nt [i]->height - t->height);
        LandVector d1 = land_vector(dx [i + 1], dy [i + 1], nt [i + 1]->height - t->height);
        LandVector normal = land_vector_cross(land_vector_normalize(d0), land_vector_normalize(d1));
        l += land_vector_dot(light, normal);
        nx += normal.x;
        ny += normal.y;
        nz += normal.z;
    }
    l /= 4.0;
    float n = sqrt(nx * nx + ny * ny + nz * nz);
    if (n > 0.01) {
        nx /= n;
        ny /= n;
        nz /= n;
    }
    else {
        nx = 0;
        ny = 0;
        nz = 1;
    }
    t->dark = 128 + l * 127;
    t->nx = nx * 100;
    t->ny = ny * 100;
    t->nz = nz * 100;
}
void isles_generate(Landscape * landscape, bool add_hills) {
    int w = landscape->w;
    int m = landscape->w - 1;
    int c = w / 2;
    int minr = 8;
    int maxr = 32;
    int count = w * w / (maxr * maxr * 2);
    if (! add_hills) {
        count = 0;
    }
    Isle isles [count];
    for (int i = 0; i < count; i += 1) {
        // starting island
        if (i == 0) {
            isles [i].x = w / 2;
            isles [i].y = w / 2;
            isles [i].radius = maxr;
        }
        else {
            isles [i].x = land_rand(maxr, m - maxr);
            isles [i].y = land_rand(maxr, m - maxr);
            isles [i].radius = land_rand(minr, maxr);
        }
        int rx = isles [i].x - c;
        int ry = isles [i].y - c;
        int rr = rx * rx + ry * ry;
        int icy = c * 0.6;
        isles [i].kind = 1;
        if (rr > icy * icy) {
            isles [i].kind = 2;
        }
        draw_isle(landscape, isles + i);
    }
    isles_update_light(landscape);
}
void isles_update_light(Landscape * landscape) {
    isles_update_normals_area(landscape, 0, 0, landscape->w, landscape->w);
}
void isles_update_normals_area(Landscape * landscape, int xp, int yp, int w, int h) {
    for (int y = yp; y < yp + h; y += 1) {
        for (int x = xp; x < xp + w; x += 1) {
            update_light(landscape, x, y);
        }
    }
}
