#ifndef _GAME_LANDSCAPE_
#define _GAME_LANDSCAPE_
#include <land/glsl.h>
typedef struct Chunk Chunk;
typedef struct Tile Tile;
typedef struct Adjustment Adjustment;
typedef struct Landscape Landscape;
#include "common.h"
#include "camera.h"
#include "isles.h"
#include "land/openglutil.h"
#include "rivers.h"
#define LIGHTMAP_RESOLUTION 1
struct Chunk {
    GLfloat * v;
    int stride;
    GLuint vbo;
    bool required : 1;
    bool filled : 1;
    bool uploaded : 1;
    bool empty : 1;
    int x, y;
    int bytes;
};
struct Tile {
    float height;
    float dark;
    float river;
    int kind;
    char nx, ny, nz;
    unsigned int resource : 4;
    unsigned int amount : 4;
    uint8_t visible;
};
enum {
    TK_WATER,
    TK_GRASS,
    TK_MOUNTAIN,
    TK_SNOW,
    TK_DESERT
};
struct Adjustment {
    float height;
    int x, y, b, s;
};
struct Landscape {
    int w /* needs to be powers of two, map is always square */;
    Tile * heightmap;
    double max_height;
    double snow_top;
    bool want_light_tops;
    bool want_rivers;
    int plateau_levels;
    int water_levels;
    int blur;
    LandImage * lightmap;
    unsigned char * lightmap_rgba;
    LandGLSLShader * shader;
    int chunks_count;
    Chunk * chunks;
    int chunk_s;
    int stride;
    int vertices_per_tile;
    int max_visible_chunks_s;
    int visible_chunks_w /* <= max_visible_chunks_s */;
    int visible_chunks_h /* <= max_visible_chunks_s */;
    LandColor river_color;
    LandColor fog_color;
    LandArray * adjustments;
    LandColor(* color_for_height)(Landscape * self, float h);
};
Landscape* landscape_create(int logsize, int seed);
Landscape* landscape_create_chunks(int logsize, int chunks, int seed);
Landscape* landscape_create_base(void);
void landscape_create_map(Landscape * self, int seed);
void landscape_create_map_flat(Landscape * self, float h);
void landscape_create_map_heightmap(Landscape * self, int seed);
void landscape_process_noise(Landscape * self, LandNoise * blur);
void landscape_calculate_light(Landscape * self);
void landscape_create_colors(Landscape * self);
#define Z(X, Y) get_height(self, x + (X), y + (Y))
Tile* landscape_get_tile(Landscape * self, int x, int y);
#define F LandFloat
float landscape_get_height_interpolated(Landscape * self, F px, F py);
void landscape_get_normal_interpolated(Landscape * self, F px, F py, F * rnx, F * rny, F * rnz);
void landscape_render(Landscape * self, Camera * cam);
void landscape_render2(Landscape * self, Camera * cam, Camera * cam2);
void landscape_draw_grid(Landscape * self, Camera * cam);
void landscape_debug_clear_chunks(Landscape * self);
void landscape_destroy(Landscape * self);
void landscape_color(Landscape * self, float x, float y, float r, float g, float b);
void landscape_height(Landscape * self, int x, int y, float height);
void landscape_invalidate(Landscape * self);
void landscape_make_circle_visible(Landscape * self, int cx, int cy, int rn, int rf);
void landscape_adjust(Landscape * ls, int x, int y, int s, int b, float height);
LandImage* landscape_create_visible_picture(Landscape * ls);
void landscape_apply_visibility_picture(Landscape * ls, LandImage * pic);
#endif
