#ifndef _GAME_MESH_
#define _GAME_MESH_
typedef struct MeshMarker MeshMarker;
typedef struct Mesh Mesh;
#include "main.h"
#include "object.h"
#include "land/util2d.h"
#include "land/glsl.h"
struct MeshMarker {
    char * name;
    Land4x4Matrix transform;
};
struct Mesh {
    int n /* the number of vertices */;
    int stride /* floats to the next vertex */;
    GLfloat * v;
    GLuint vbo;
    LandArray * markers /* MeshMarker */;
    LandImage * texture;
    LandCSGAABB aabb;
};
Mesh* mesh_new(void);
Mesh* mesh_new_textured(void);
MeshMarker* mesh_find_marker(Mesh * self, char const * name);
void mesh_destroy(Mesh * self);
void mesh_add_marker(Mesh * self, Object * o);
void mesh_add_object(Mesh * self, Object * o);
void mesh_get_bounds(Mesh * self, LandFloat * x1, LandFloat * y1, LandFloat * z1, LandFloat * x2, LandFloat * y2, LandFloat * z2);
void mesh_get_bounds_if_less(Mesh * self, LandFloat * x1, LandFloat * y1, LandFloat * z1, LandFloat * x2, LandFloat * y2, LandFloat * z2, LandFloat x, LandFloat y, LandFloat z);
void mesh_get_bounds_if_between(Mesh * self, LandFloat * x1, LandFloat * y1, LandFloat * z1, LandFloat * x2, LandFloat * y2, LandFloat * z2, LandFloat xl, LandFloat yl, LandFloat zl, LandFloat xg, LandFloat yg, LandFloat zg);
void mesh_set_origin(Mesh * self, float x, float y, float z);
void mesh_recalculate_normals(Mesh * self);
void mesh_triangle_callback(Mesh * self, void(* callback)(LandVector t0, LandVector t1, LandVector t2, void * data), void * data);
int mesh_write_frames(LandFile * f, LandArray * frames);
LandArray* mesh_read_frames(LandFile * f);
void mesh_transform_positions(Mesh * self, Land4x4Matrix * matrix);
#endif
