#ifndef _GAME_OBJECT_
#define _GAME_OBJECT_
typedef enum Dummy Dummy;
typedef struct Object Object;
#include "common.h"
#include "land/csg/csg.h"
#include "material.h"
#include "land/csg/csg_shapes.h"
enum Dummy {
    DummyBegin=1,
    DummyUnion=2,
    DummyTransparent=4,
    DummySeparate=8,
    DummyMerge=16,
    DummyRecalculateSmooth=32,
    DummyRecalculateFace=64
};
struct Object {
    LandCSG * csg;
    bool inverse;
    bool is_marker;
    int dummy;
    char * name /* for marker */;
    Land4x4Matrix transform /* for marker */;
};
Object* object_new(void);
Object* object_clone(Object * original);
void object_destroy(Object * self);
Object* object_sphere(int slices, int rings, Material * m);
Object* object_spherez(int slices, int rings, Material * m);
Object* object_dome(bool open, bool flipped, int slices, int discs, Material * m);
Object* object_hemi(bool open, int slices, int discs, Material * m);
Object* object_tetrasphere(int divisions, Material * m);
Object* object_icosphere(int divisions, Material * m);
Object* object_cylinder(int segments, int discs, Material * m);
Object* object_cylinder_open(int segments, int discs, Material * m);
Object* object_cut_cone(int segments, float radius, Material * m);
Object* object_cone(int segments, Material * m);
Object* object_n_pyramid(int n, Material * m);
Object* object_cut_pyramid(float x, float y, Material * m);
Object* object_box(Material * m);
Object* object_block(int x, int y, int z, Material * m);
Object* object_grid(int x, int y, Material * m);
Object* object_regular(int n, Material * m);
Object* object_pyramid(Material * m);
Object* object_pyramid2(float x, float y, float z, Material * m);
Object* object_irregular_pyramid(float x, float y, float z, Material * m);
Object* object_irregular_cut_pyramid(float x, float y, float z, float w, float h, Material * m);
Object* object_torus(int slices, int rings, LandFloat diameter, Material * m);
Object* object_prism(int n, Material * m);
Object* object_prism2(LandFloat ax, LandFloat ay, LandFloat az, LandFloat bx, LandFloat by, LandFloat bz, LandFloat d, Material * m);
Object* object_trapezoid(LandFloat x1, LandFloat x2, Material * m);
Object* object_tetra(LandFloat ax, LandFloat ay, LandFloat az, LandFloat bx, LandFloat by, LandFloat bz, LandFloat cx, LandFloat cy, LandFloat cz, LandFloat dx, LandFloat dy, LandFloat dz, Material * m);
void object_subdivide(Object * o, Material * m);
void object_shrinkwrap(Object * o, Material * m, int subdiv);
Object* object_mark(char const * name);
void object_transform(Object * o, Land4x4Matrix * matrix);
void object_union(Object * self, Object * other);
char* object_desc(Object * o);
#endif
