#include "game/rivers.h"
#define _scramble_min(x, y) ((y) < (x) ? (y) : (x))
#define _scramble_max(x, y) ((y) > (x) ? (y) : (x))
static int _get(Rivers * r, int * a, int x, int y);
static bool _move(Rivers * r, int * x, int * y);
static bool _move_reverse(Rivers * r, int * x, int * y);
static int _endcomp(void const * a, void const * b);
static int _cb(int rx, int ry, void * data);
static bool _find_closest_under(LandNoise * noise, float h, int * x, int * y);
static void _paint(LandNoise * noise, int x, int y);
static int _get(Rivers * r, int * a, int x, int y) {
    return a [((y + r->h) % r->h) * r->w + ((x + r->w) % r->w)];
}
static bool _move(Rivers * r, int * x, int * y) {
    float v = land_noise_at(r->noise, * x, * y);
    if (v <= r->water_height) {
        return 0;
    }
    float a = land_noise_at(r->noise, * x + 1, * y);
    float b = land_noise_at(r->noise, * x, * y + 1);
    float c = land_noise_at(r->noise, * x - 1, * y);
    float d = land_noise_at(r->noise, * x, * y - 1);
    float mab = _scramble_min(a, b);
    float mcd = _scramble_min(c, d);
    float m = _scramble_min(mab, mcd);
    if (m >= v) {
        return 0;
    }
    if (a == m) {
        (* x)++;
    }
    else if (b == m) {
        (* y)++;
    }
    else if (c == m) {
        (* x)--;
    }
    else if (d == m) {
        (* y)--;
    }
    else {
        return 0;
    }
    return 1;
}
static bool _move_reverse(Rivers * r, int * x, int * y) {
    //int av = _get(r, r.add, *x, *y)
    int aa = _get(r, r->add, * x + 1, * y);
    int ab = _get(r, r->add, * x, * y + 1);
    int ac = _get(r, r->add, * x - 1, * y);
    int ad = _get(r, r->add, * x, * y - 1);
    float v = land_noise_at(r->noise, * x, * y);
    float a = land_noise_at(r->noise, * x + 1, * y);
    float b = land_noise_at(r->noise, * x, * y + 1);
    float c = land_noise_at(r->noise, * x - 1, * y);
    float d = land_noise_at(r->noise, * x, * y - 1);
    if (v >= r->snow_height) {
        return 0;
    }
    if (a < v) {
        aa = 0;
    }
    if (b < v) {
        ab = 0;
    }
    if (c < v) {
        ac = 0;
    }
    if (d < v) {
        ad = 0;
    }
    int maaab = _scramble_max(aa, ab);
    int macad = _scramble_max(ac, ad);
    int am = _scramble_max(maaab, macad);
    if (am == 0) {
        return 0;
    }
    if (am == aa) {
        (* x)++;
    }
    else if (am == ab) {
        (* y)++;
    }
    else if (am == ac) {
        (* x)--;
    }
    else if (am == ad) {
        (* y)--;
    }
    else {
        return 1;
    }
    return 1;
}
static int _endcomp(void const * a, void const * b) {
    End * ae = * (End * *) a;
    End * be = * (End * *) b;
    if (ae->volume > be->volume) {
        return - 1;
    }
    if (ae->volume < be->volume) {
        return 1;
    }
    return 0;
}
static int _cb(int rx, int ry, void * data) {
    Data * _data = data;
    float v = land_noise_at(_data->noise, rx, ry);
    if (v < _data->h) {
        _data->x = rx;
        _data->y = ry;
        return 1;
    }
    return 0;
}
static bool _find_closest_under(LandNoise * noise, float h, int * x, int * y) {
    Data data = {noise, h, * x, * y};
    int r = helper_diamond_callback(* x, * y, 0, 19, _cb, & data);
    if (r) {
        * x = data.x;
        * y = data.y;
        return 1;
    }
    return 0;
}
LandNoise* rivers(LandNoise * noise, float water_height, float snow_height) {
    int w = noise->w;
    int h = noise->h;
    Rivers r;
    double t0 = land_get_time();
    LandNoise * count = land_noise_new(LandNoiseValue, 0);
    land_noise_set_size(count, w, h);
    land_noise_prepare(count);
    r.noise = noise;
    r.count = count;
    r.add = land_calloc(w * h * sizeof (* r.add));
    r.end = land_calloc(w * h * sizeof (* r.end));
    r.w = w;
    r.h = h;
    r.snow_height = snow_height;
    r.water_height = water_height;
    for (int y = 0; y < h; y += 4) {
        for (int x = 0; x < w; x += 4) {
            int rx = x;
            int ry = y;
            float v = land_noise_at(noise, rx, ry);
            if (v < snow_height) {
                continue;
            }
            r.add [rx + w * ry]++;
            int l = 0;
            while (1) {
                while (_move(& r, & rx, & ry)) {
                    rx = (rx + w) % w;
                    ry = (ry + h) % h;
                    r.add [rx + w * ry]++;
                    l += 1;
                }
                r.end [rx + w * ry] += l;
                v = land_noise_at(noise, rx, ry);
                if (v <= water_height) {
                    break;
                }
                int tx = rx;
                int ty = ry;
                if (! _find_closest_under(noise, v, & tx, & ty)) {
                    break;
                }
                rx = (tx + w) % w;
                ry = (ty + h) % h;
            }
        }
    }
    print("path: %.2f", land_get_time() - t0);
    LandArray * ends = land_array_new();
    for (int y = 0; y < h; y += 1) {
        for (int x = 0; x < w; x += 1) {
            if (r.end [x + w * y] == 0) {
                continue;
            }
            End * end;
            land_alloc(end);
            end->x = x;
            end->y = y;
            end->volume = r.end [x + w * y];
            land_array_add(ends, end);
        }
    }
    land_array_sort(ends, _endcomp);
    int e = 0;
    //printf("rivers:\n")
    {
        LandArrayIterator __iter0__ = LandArrayIterator_first(ends);
        for (End * end = LandArrayIterator_item(ends, &__iter0__); LandArrayIterator_next(ends, &__iter0__); end = LandArrayIterator_item(ends, &__iter0__)) {
            if (end->volume > 1000 || e < 10) {
                int rx = end->x;
                int ry = end->y;
                while (_move_reverse(& r, & rx, & ry)) {
                    rx = (rx + w) % w;
                    ry = (ry + h) % h;
                    end->length++;
                }
            }
            e++;
        }
    }
    {
        LandArrayIterator __iter0__ = LandArrayIterator_first(ends);
        for (End * end = LandArrayIterator_item(ends, &__iter0__); LandArrayIterator_next(ends, &__iter0__); end = LandArrayIterator_item(ends, &__iter0__)) {
            if (end->length < 10) {
                continue;
            }
            //printf("%d %d %d %d\n", end.x, end.y, end.volume, end.length)
            int rx = end->x;
            int ry = end->y;
            while (_move_reverse(& r, & rx, & ry)) {
                rx = (rx + w) % w;
                ry = (ry + h) % h;
                r.add [rx + w * ry] = 0;
                _paint(count, rx, ry);
            }
        }
    }
    land_free(r.add);
    land_free(r.end);
    return count;
}
static void _paint(LandNoise * noise, int x, int y) {
    int r = 5;
    for (int v = - r; v < r + 1; v += 1) {
        for (int u = - r; u < r + 1; u += 1) {
            int dd = u * u + v * v;
            if (dd > r * r) {
                continue;
            }
            int rx = (x + u + noise->w) % noise->w;
            int ry = (y + v + noise->h) % noise->h;
            float a = noise->cache [rx + ry * noise->w];
            float b = r - sqrt(dd);
            b /= r;
            b = pow(b, 2) / 2;
            a = _scramble_max(a, b);
            noise->cache [rx + ry * noise->w] = a;
        }
    }
}
