#include "script/helper.h"
void helper_add_fnc(lua_State * l, char const * name, void * fun) {
    lua_getglobal(l, name);
    bool already = ! lua_isnil(l, - 1);
    lua_pop(l, 1);
    if (already) {
        return ;
    }
    if (DebugLogLua) {
        print("registering function %s", name);
    }
    lua_register(l, name, fun);
    PerThread * pt = script_get_per_thread(l);
    land_hash_insert(pt->registered_functions, name, fun);
}
// table must be top of stack
// stack is not modified after the call
int helper_get_table_int(lua_State * l, str name) {
    lua_pushstring(l, name) /* [name] */;
    lua_rawget(l, - 2) /* [x] */;
    int x = lua_tointeger(l, - 1);
    lua_pop(l, 1) /* [] */;
    return x;
}
lua_Number helper_get_table_float(lua_State * l, str name) {
    lua_pushstring(l, name) /* [name] */;
    lua_rawget(l, - 2) /* [x] */;
    lua_Number x = lua_tonumber(l, - 1);
    lua_pop(l, 1) /* [] */;
    return x;
}
// this one replaces the table with the value
void helper_get_table_value(lua_State * l, str name) {
    lua_pushstring(l, name) /* [table name] */;
    lua_rawget(l, - 2) /* [table value] */;
    lua_remove(l, - 2) /* [value] */;
}
    /* Returns an array filled with allocated strings. The array and the
     * strings should be freed by the caller when done.
     */
LandArray* helper_get_table_keys(lua_State * l) {
    // [table] must be on the stack (and is removed)
    LandArray * a = land_array_new();
    if (lua_isnil(l, - 1)) {
        return a;
    }
    lua_pushnil(l) /* [table nil] */;
    while (lua_next(l, - 2)) {
        // [table key val]
        lua_pop(l, 1) /* [table key] */;
        str v = lua_tostring(l, - 1);
        if (v) {
            char * key = land_strdup(v);
            land_array_add(a, key);
        }
    }
    lua_pop(l, 1) /* [] */;
    return a;
}
