#include "unit.h"
#include "game.h"
Unit* unit_new(int kind, float x, float y) {
    Application * app = app_global();
    int hx, hy, ix, iy;
    pos_to_hex(x, y, & hx, & hy);
    hex_to_int(hx, hy, & ix, & iy);
    float z = get_top_z(ix, iy);
    Unit * self;
    land_alloc(self);
    self->kind = kind;
    land_array_add(app->units, self);
    camera_init(& self->cam);
    self->cam.p.x = ix;
    self->cam.p.y = iy;
    self->cam.p.z = z;
    camera_look_into(& self->cam, - x, - y);
    camera_change_freely(& self->cam, 0, 0, land_rnd(- pi / 8, pi / 8));
    self->speed = 1;
    self->start = self->cam.p;
    return self;
}
void units_clear(void) {
    Application * app = app_global();
    {
        LandArrayIterator __iter0__ = LandArrayIterator_first(app->units);
        for (Unit * unit = LandArrayIterator_item(app->units, &__iter0__); LandArrayIterator_next(app->units, &__iter0__); unit = LandArrayIterator_item(app->units, &__iter0__)) {
            land_free(unit);
        }
    }
    land_array_clear(app->units);
}
void units_tick(void) {
    Application * app = app_global();
    {
        LandArrayIterator __iter0__ = LandArrayIterator_first(app->units);
        for (Unit * unit = LandArrayIterator_item(app->units, &__iter0__); LandArrayIterator_next(app->units, &__iter0__); unit = LandArrayIterator_item(app->units, &__iter0__)) {
            if (unit->dead) {
                continue;
            }
            if (unit->kind == UnitAllefant) {
                unit_tick(unit);
            }
            if (unit->kind == UnitMothership) {
                unit_fly(unit);
            }
        }
    }
    if (app->ticks % 60 == 0) {
        {
            LandArrayIterator __iter0__ = LandArrayIterator_first(app->units);
            for (Unit * unit = LandArrayIterator_item(app->units, &__iter0__); LandArrayIterator_next(app->units, &__iter0__); unit = LandArrayIterator_item(app->units, &__iter0__)) {
                if (unit->dead) {
                    continue;
                }
                unit_plan(unit);
            }
        }
    }
}
void catch(Unit * self) {
    Application * app = app_global();
    LandVector * p = & self->cam.p;
    // get screen coordinates of elephant
    LandVector s = get_point(& app->cam, p->x, p->y, p->z);
    // get 3d position under it
    LandVector pos = world_get_position_under_z(& app->cam, s.x, s.y);
    int hx, hy;
    pos_to_hex(pos.x, pos.y, & hx, & hy);
    for (int y = - 1; y < 2; y += 1) {
        for (int x = - 1; x < 2; x += 1) {
            int z = find_highest_z(hx + x, hy + y);
            int bz = check_bridge_z(hx + x, hy + y, z);
            if (bz > 0) {
                int ix, iy;
                hex_to_int(hx + x, hy + y, & ix, & iy);
                self->fall = 0;
                self->bridge = 1;
                self->walk = 0;
                p->z = bz;
                p->x = ix;
                p->y = iy;
                return ;
            }
        }
    }
}
void unit_plan(Unit * self) {
    Application * app = app_global();
    self->left = land_rand(0, 1);
    if (self->kind == UnitAllefant) {
        LandVector * p = & self->cam.p;
        int hx, hy;
        pos_to_hex(p->x, p->y, & hx, & hy);
        int bz = check_bridge_z(hx, hy, floor(p->z / 4));
        if (bz <= 0) {
            self->bridge = 0;
        }
        if (p->z < - 200) {
            self->dead = 1;
            app->died++;
        }
        if (self->fall) {
            catch(self);
        }
        float dx = rwrap(p->x - app->goal_x, app->world->landscape->w);
        float dy = rwrap(p->y - app->goal_y, app->world->landscape->w);
        if (dx * dx + dy * dy < 4 * 4) {
            self->dead = 1;
            app->retrieved++;
            land_sound_play(app->win, 1, 0, 1);
            return ;
        }
        if (self->bridge) {
            int d = get_hex_dir(- self->cam.y.x, - self->cam.y.y);
            bool found = 0;
            for (int i = 0; i < 6; i += 1) {
                if (i != 3) {
                    int x2 = hx;
                    int y2 = hy;
                    int z2 = p->z / 4;
                    add_dir(& x2, & y2, d);
                    if (check_bridge_z(x2, y2, z2) > 0) {
                        int nx, ny;
                        hex_to_int(x2, y2, & nx, & ny);
                        camera_look_into(& self->cam, nx - p->x, ny - p->y);
                        found = 1;
                        break;
                    }
                }
                if (self->left) {
                    d = (d + 5) % 6;
                }
                else {
                    d = (d + 1) % 6;
                }
            }
            if (! found) {
                int x2 = hx;
                int y2 = hy;
                add_dir(& x2, & y2, d);
                int nx, ny;
                hex_to_int(x2, y2, & nx, & ny);
                camera_look_into(& self->cam, nx - p->x, ny - p->y);
            }
        }
    }
    if (self->kind == UnitMothership) {
        LandVector d = land_vector_sub(self->start, self->cam.p);
        self->drift.x = land_rnd(- 0.01, 0.01);
        self->drift.z = land_rnd(- 0.01, 0.01);
        if (self->cam.x.z > .1) {
            self->drift.y = - 0.01;
        }
        if (self->cam.x.z < - .1) {
            self->drift.y = 0.01;
        }
        if (self->cam.y.z > .1) {
            self->drift.x = 0.01;
        }
        if (self->cam.y.z < - .1) {
            self->drift.x = - 0.01;
        }
        if (d.z > 4) {
            self->drift.x = 0.01;
        }
        if (d.z < - 4) {
            self->drift.x = - 0.01;
        }
        if (land_vector_norm(d) > 40) {
            if (land_cross2d(d.x, d.y, self->cam.y.x, self->cam.y.y) > 0) {
                self->drift.z = - 0.03;
            }
            else {
                self->drift.z = 0.03;
            }
        }
        if (app->release) {
            self->counter++;
            if (self->counter == 5) {
                self->counter -= 5;
                if (app->released < app->elephants) {
                    app->released++;
                    Unit * u = unit_new(UnitAllefant, self->start.x, self->start.y);
                    float dx = rwrap(app->goal_x - u->cam.p.x, app->world->landscape->w);
                    float dy = rwrap(app->goal_y - u->cam.p.y, app->world->landscape->w);
                    camera_look_into(& u->cam, dx, dy);
                    u->cam.p.z = 30;
                }
            }
        }
    }
}
void unit_fly(Unit * self) {
    camera_change_freely(& self->cam, self->drift.x, self->drift.y, self->drift.z);
    self->drift.x *= 0.9;
    self->drift.y *= 0.9;
    self->drift.z *= 0.9;
    float s = - 0.1;
    land_vector_iadd(& self->cam.p, land_vector_mul(self->cam.y, s));
    if (self->visible < 1) {
        self->visible += 0.01;
    }
}
void unit_tick(Unit * self) {
    Application * app = app_global();
    Landscape * ls = app->world->landscape;
    LandVector back = self->cam.y;
    LandVector * p = & self->cam.p;
    LandVector old_p = * p;
    LandVector v = land_vector_mul(back, - 0.1 * self->speed);
    if (self->fall) {
        land_vector_iadd(p, land_vector(0, 0, - 0.2));
        camera_change_freely(& self->cam, - 0.03, 0, 0);
    }
    else if (self->walk) {
        land_vector_iadd(p, v);
        bool collide = 0;
        int hx, hy;
        pos_to_hex(p->x, p->y, & hx, & hy);
        int wz = check_wall_z(hx, hy, floor(p->z / 4));
        float z, z2;
        if (wz > 0) {
            if (wz > p->z + 0.25) {
                collide = 1;
            }
            else {
                z = z2 = wz;
            }
        }
        else {
            z = landscape_get_height_interpolated(ls, p->x, p->y);
            z2 = landscape_get_height_interpolated(ls, p->x - back.x, p->y - back.y);
        }
        int bz = check_bridge_z(hx, hy, floor(p->z / 4));
        bool bridge = 0;
        if (bz > 0) {
            if (p->z > bz - 1 && p->z < bz + 1) {
                bridge = 1;
                self->bridge = 1;
                z = z2 = bz;
            }
        }
        if (! bridge && self->bridge) {
            if (z == 0 || z < p->z - 1) {
                collide = 1;
            }
            else {
                self->bridge = 0;
            }
        }
        if (collide) {
            * p = old_p;
            camera_change_freely(& self->cam, 0, 0, self->left ? pi / (- 32) : pi / 32);
        }
        else {
            if (z > p->z - 0.25) {
                p->z = z;
            }
            else {
                p->z -= 0.25;
            }
            camera_look_into(& self->cam, v.x, v.y);
            float gradient = atan2(z2 - z, 1);
            camera_change_freely(& self->cam, gradient, 0, 0);
            self->speed = 1;
            if (z2 > z) {
                self->speed = 1 + z - z2;
                if (self->speed < 0.1) {
                    self->speed = 0.1;
                }
            }
        }
    }
    else {
        float z = get_top_z(self->cam.p.x, self->cam.p.y);
        if (z < 0.1) {
            self->fall = 1;
            land_sound_play(app->yell, 1, 0, 2);
        }
        if (self->cam.p.z > z + 0.25) {
            self->cam.p.z -= 0.25;
        }
        else {
            self->walk = 1;
        }
    }
    //LandVector pos0 = self.cam.p
    self->cam.p.x = wrap(self->cam.p.x, ls->w);
    self->cam.p.y = wrap(self->cam.p.y, ls->w);
    //pos0.z = 0
    //if p.x * p.x + p.y * p.y > 256 * 256 and land_vector_dot(self.cam.y, pos0) < 0:
    //    camera_look_into(&self.cam, -p.x, -p.y)
    //    camera_change_freely(&self.cam, 0, 0, land_rnd(-pi / 8, pi / 8))
    Tile * tile = landscape_get_tile(ls, floor(p->x), floor(p->y));
    self->visible_want = tile->visible / 255.0;
    if (self->visible > self->visible_want) {
        self->visible -= 0.01;
    }
    if (self->visible < self->visible_want) {
        self->visible += 0.01;
    }
    if (! self->bridge && tile->height < 0.1) {
        if (! self->fall) {
            self->fall = 1;
            land_sound_play(app->yell, 1, 0, 2);
        }
    }
}
void units_draw_offset(float ox, float oy) {
    Application * app = app_global();
    LandVector light = app->light;
    {
        LandArrayIterator __iter0__ = LandArrayIterator_first(app->units);
        for (Unit * unit = LandArrayIterator_item(app->units, &__iter0__); LandArrayIterator_next(app->units, &__iter0__); unit = LandArrayIterator_item(app->units, &__iter0__)) {
            if (unit->dead) {
                continue;
            }
            LandVector backup = unit->cam.p;
            unit->cam.p.x += ox;
            unit->cam.p.y += oy;
            unit_draw(unit, light);
            unit->cam.p = backup;
        }
    }
    model_render_done();
}
void unit_draw(Unit * self, LandVector light) {
    Application * app = app_global();
    Model3d * model = NULL;
    if (self->kind == UnitAllefant) {
        model = app->model_allefant;
    }
    if (self->kind == UnitMothership) {
        model = app->model_mothership;
    }
    if (model) {
        draw_model(model, & self->cam, self->visible, light);
    }
}
void draw_model(Model3d * c, Camera * cam, float alpha, LandVector light) {
    Application * app = app_global();
    if (! c) {
        return ;
    }
    if (! c->frames) {
        return ;
    }
    int f = app->ticks / 4;
    int n = land_array_count(c->frames);
    if (n == 5) {
        f %= 8;
        if (f >= 5) {
            f = 8 - f;
        }
    }
    f %= n;
    Mesh * mesh = land_array_get_nth(c->frames, f);
    Land4x4Matrix mv = camera_matrix(& app->cam);
    Land4x4Matrix mc = camera_forward_matrix(cam);
    float s = c->scale;
    mc = land_4x4_matrix_mul(mc, land_4x4_matrix_scale(s, s, s));
    Land4x4Matrix matrix = mv;
    matrix = land_4x4_matrix_mul(matrix, mc);
    if (mesh) {
        model_render_object_light3(mesh, matrix, world_get_projection_matrix(), cam, light, alpha);
    }
}
